import inspect
from dataclasses import dataclass, field, fields, is_dataclass, make_dataclass
from functools import wraps
from typing import Any, Callable, Type, TypeVar

from omegaconf import DictConfig, ListConfig, OmegaConf
from typing_extensions import dataclass_transform

from .traversal import traverse
from .types import get_type
from .utils import SpringsWarnings

DT = TypeVar("DT")


class FlexyClass:
    ...


@dataclass_transform()
def make_flexy(cls_: Type[DT]) -> Type[DT]:
    """A flexyclass is like a dataclass, but it supports partial
    specification of properties."""

    if not inspect.isclass(cls_) or not is_dataclass(cls_):
        raise TypeError(f"flexyclass must decorate a dataclass, not {cls_}")

    # type ignore is for pylance, which freaks out a bit otherwise
    new_cls: Type[DT] = type(
        f"FlexyClass{cls_.__name__}", (cls_, FlexyClass), {}  # type: ignore
    )

    return new_cls


@dataclass_transform()
def flexyclass(cls: Type[DT]) -> Type[DT]:  # type: ignore
    """A flexyclass is like a dataclass, but it supports partial
    specification of properties."""
    SpringsWarnings.flexyclass()
    return make_flexy(dataclass(cls))  # type: ignore


def flexy_field(type_: Type[DT], /, **kwargs: Any) -> DT:
    """A flexy_ field is like dataclass.field, but it supports
    passing arbitrary keyword arguments to a flexyclass.

    Args:
        type_: The flexyclass this field is for.
        kwargs: Any keyword arguments to pass to the flexyclass.
    """
    SpringsWarnings.flexyfield()

    if not issubclass(type_, FlexyClass) and not is_dataclass(type_):
        raise TypeError(f"flexy_field must receive a flexyclass, not {type_}")

    # find the argument that are extra from what has been defined in
    # the flexyclass.
    known_fields = set(f.name for f in fields(type_))
    known_kwargs, extra_kwargs = {}, {}
    for k, v in kwargs.items():
        if k in known_fields:
            known_kwargs[k] = v
        else:
            extra_kwargs[k] = v

    # decorate
    if extra_kwargs:
        cls_ = make_dataclass(
            cls_name=f"{type_.__name__}_{'_'.join(extra_kwargs)}",
            fields=[
                (f_name, type(f_value), field(default_factory=lambda: f_value))
                for f_name, f_value in extra_kwargs.items()
            ],
            bases=(type_,),
        )
    else:
        cls_ = type_

    new_field: DT = field(
        default_factory=lambda: cls_(**known_kwargs)  # type: ignore
    )
    return new_field


DictOrListConfig = TypeVar("DictOrListConfig", DictConfig, ListConfig)


def unlock_all_flexyclasses(
    cast_fn: Callable[..., DictOrListConfig]
) -> Callable[..., DictOrListConfig]:
    @wraps(cast_fn)
    def unlock_fn(*args, **kwargs):
        config_node = cast_fn(*args, **kwargs)

        for spec in traverse(
            node=config_node, include_nodes=True, include_root=True
        ):

            if not isinstance(spec.value, DictConfig):
                # only DictConfigs can be flexyclasses
                continue

            typ_ = get_type(spec.value)
            if typ_ and issubclass(typ_, FlexyClass):
                OmegaConf.set_struct(spec.value, False)

        return config_node

    return unlock_fn



import logging
from multiprocessing import current_process
from pathlib import Path
from typing import Optional, Type, Union

from .utils import SpringsWarnings


class configure_logging:
    _DEBUG_MODE: bool = False

    @classmethod
    def debug(
        cls: Type["configure_logging"], *args, **kwargs
    ) -> logging.Logger:

        if "logging_level" in kwargs:
            SpringsWarnings.argument("logging_level", "debug")

        kwargs["logging_level"] = logging.DEBUG

        cls._DEBUG_MODE = True

        return cls.__new__(cls, *args, **kwargs)

    def __new__(  # type: ignore
        cls: Type["configure_logging"],
        logger_name: Optional[str] = None,
        file_logging_path: Optional[Path] = None,
        make_dir_if_missing: bool = True,
        fmt: str = "[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
        datefmt: str = "%Y-%m-%d %H:%M:%S",
        logging_level: Union[int, str] = logging.INFO,
        force_root_reattach: bool = True,
        root_formatter_kwargs: Optional[dict] = None,
        stream_handler_kwargs: Optional[dict] = None,
        file_handler_kwargs: Optional[dict] = None,
        basic_config_kwargs: Optional[dict] = None,
        additional_handlers: Optional[list] = None,
    ) -> logging.Logger:
        """A big function that speeds up configuration of logging"""

        if isinstance(logging_level, str):
            logging_level = getattr(logging, logging_level)

        # we abide by the global debugging flag
        logging_level = logging.DEBUG if cls._DEBUG_MODE else logging_level

        # change how the formatter looks
        kw = root_formatter_kwargs or {}
        root_formatter = logging.Formatter(fmt=fmt, datefmt=datefmt, **kw)

        handlers = list(additional_handlers or [])

        kw = stream_handler_kwargs or {}
        root_stream_handler = logging.StreamHandler(**kw)
        root_stream_handler.setFormatter(root_formatter)
        handlers.append(root_stream_handler)

        if file_logging_path is not None:
            file_logging_path = Path(file_logging_path)

            if make_dir_if_missing:
                file_logging_path.parent.mkdir(parents=True, exist_ok=True)

            path_name = (
                f"{file_logging_path.stem}"
                f"_{current_process().name}"
                f"{file_logging_path.suffix}"
            )

            kw = file_handler_kwargs or {}
            root_file_handler = logging.FileHandler(
                filename=file_logging_path.parent / path_name, **kw
            )
            root_file_handler.setFormatter(root_formatter)
            handlers.append(root_file_handler)

        # add the handlers to the root logger, force reattaching other loggers
        kw = basic_config_kwargs or {}
        logging.basicConfig(
            level=logging_level,
            force=force_root_reattach,
            handlers=handlers,
            **kw,
        )
        return logging.getLogger(logger_name)



import os
import sys
from argparse import Action, ArgumentParser
from dataclasses import dataclass, fields, is_dataclass
from inspect import getfile, getfullargspec, isclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from omegaconf import MISSING, DictConfig
from typing_extensions import Concatenate, ParamSpec

from .core import (
    from_dataclass,
    from_file,
    from_none,
    from_options,
    merge,
    traverse,
    validate,
)
from .utils import PrintUtils

# parameters for the main function
MP = ParamSpec("MP")

# type for the configuration
CT = TypeVar("CT")

# return type for main function
RT = TypeVar("RT")


@dataclass
class Flag:
    name: str
    help: str
    action: str = MISSING
    default: Optional[Any] = MISSING
    nargs: Optional[Union[str, int]] = MISSING
    metavar: Optional[str] = MISSING
    usage_extras: Optional[str] = MISSING

    @property
    def short(self) -> str:
        return f"-{self.name[0]}"

    @property
    def usage(self) -> str:
        extras = (
            "" if self.usage_extras is MISSING else f" {self.usage_extras}"
        )
        return f"{{{self}{extras}}}"

    @property
    def long(self) -> str:
        return f"--{self.name}"

    def add_argparse(self, parser: ArgumentParser) -> Action:
        kwargs: Dict[str, Any] = {"help": self.help}
        if self.action is not MISSING:
            kwargs["action"] = self.action
        if self.default is not MISSING:
            kwargs["default"] = self.default
        if self.nargs is not MISSING:
            kwargs["nargs"] = self.nargs
        if self.metavar is not MISSING:
            kwargs["metavar"] = self.metavar

        return parser.add_argument(self.short, self.long, **kwargs)

    def __str__(self) -> str:
        return f"{self.short}/{self.long}"


@dataclass
class CliFlags:
    config: Flag = Flag(
        name="config",
        help="A path to a YAML file containing a configuration.",
        default=[],
        nargs="*",
        metavar="/path/to/yaml",
    )
    options: Flag = Flag(
        name="options",
        help="Print all default options and CLI flags.",
        action="store_true",
    )
    inputs: Flag = Flag(
        name="inputs",
        help="Print the input configuration.",
        action="store_true",
    )
    parsed: Flag = Flag(
        name="parsed",
        help="Print the parsed configuration.",
        action="store_true",
    )
    debug: Flag = Flag(
        name="debug",
        help="Enter debug mode by setting global logging to DEBUG.",
        action="store_true",
    )
    quiet: Flag = Flag(
        name="quiet",
        help="If provided, it does not print the configuration when running.",
        action="store_true",
    )
    resolvers: Flag = Flag(
        name="resolvers",
        help=(
            "Print all registered resolvers in OmegaConf, "
            "Springs, and current codebase."
        ),
        action="store_true",
    )
    nicknames: Flag = Flag(
        name="nicknames",
        help="Print all registered nicknames in Springs",
        action="store_true",
    )

    @property
    def flags(self) -> Iterable[Flag]:
        for f in fields(self):
            maybe_flag = getattr(self, f.name)
            if isinstance(maybe_flag, Flag):
                yield maybe_flag

    def add_argparse(self, parser: ArgumentParser) -> Sequence[Action]:
        return [flag.add_argparse(parser) for flag in self.flags]

    @property
    def usage(self) -> str:
        """Print the usage string for the CLI flags."""
        return " ".join(flag.usage for flag in self.flags)

    def make_cli(self, func: Callable, name: str) -> ArgumentParser:
        """Sets up argument parser ahead of running the CLI. This includes
        creating a help message, and adding a series of flags."""

        # we find the path to the script we are decorating with the
        # cli so that we can display that to the user.
        current_dir = Path(os.getcwd())
        path_to_fn_file = Path(getfile(func))
        rel_fn_file_path = str(path_to_fn_file).replace(str(current_dir), "")

        # Program name and usage added here.
        ap = ArgumentParser(
            description=f"Parser for configuration {name}",
            usage=(
                f"python3 {rel_fn_file_path} {self.usage} "
                "param1=value1 … paramN=valueN"
            ),
        )
        self.add_argparse(ap)
        return ap


def check_if_callable_can_be_decorated(func: Callable):
    expected_args = getfullargspec(func).args
    if len(expected_args) == 0:
        msg = (
            f"Function `{func.__name__}` cannot be decorated "
            f"by `config_to_program` because it does "
            f"not accept any argument."
        )
        raise RuntimeError(msg)
    elif len(expected_args) > 1:
        msg = (
            f"Function `{func.__name__}` cannot be decorated by "
            f" `config_to_program` because it expects "
            f"{len(expected_args)} > 1; If you want to pass extra "
            f"arguments to this function, use kwargs with default values."
        )
        raise RuntimeError(msg)


def check_if_valid_main_args(func: Callable, args: Sequence[Any]):
    if len(args):
        msg = (
            f"After decorating `{func.__name__}` with "
            f"`config_to_program`, do not provide any additional "
            f"arguments while invoking it; any additional parameter "
            f"should be passed as a keyword argument."
        )
        raise RuntimeError(msg)


def wrap_main_method(
    func: Callable[Concatenate[Any, MP], RT],
    name: str,
    config_node: DictConfig,
    print_fn: Optional[Callable] = None,
    *args: MP.args,
    **kwargs: MP.kwargs,
) -> RT:

    if not isinstance(config_node, DictConfig):
        raise TypeError("Config node must be a DictConfig")

    # Making sure I can decorate this function
    check_if_callable_can_be_decorated(func=func)
    check_if_valid_main_args(func=func, args=args)

    # Get argument parser and arguments
    ap = CliFlags().make_cli(func=func, name=name)
    opts, leftover_args = ap.parse_known_args()

    # setup debug
    if opts.debug:
        # relative import here not to mess things up
        from .logging import configure_logging

        configure_logging.debug()

    # Setup an utility to deal with printing
    pu = PrintUtils(print_fn=print_fn)

    # We don't run the main program if the user
    # has requested to print the any of the config.
    do_no_run = (
        opts.options
        or opts.inputs
        or opts.parsed
        or opts.resolvers
        or opts.nicknames
    )

    if opts.resolvers:
        # relative import here not to mess things up
        from .resolvers import all_resolvers

        pu.print("RESOLVERS:", *all_resolvers(), level_up=1)

    if opts.nicknames:
        from .nicknames import NicknameRegistry

        pu.print(
            "NICKNAMES:",
            *(": ".join(e) for e in NicknameRegistry.all()),
            level_up=1,
        )

    # Print default options if requested py the user
    if opts.options:
        params = traverse(config_node)

        cli_opts_repr = ("OPTS/CLI FLAG:",) + tuple(
            f"{p.path} = " + (str(p.value) if p.value != "" else "''")
            for p in params
        )
        pu.print(*cli_opts_repr, level_up=1)

    # load options from one or more config files;
    # if multiple config files are provided,
    # the latter ones can override the former ones.
    file_config = from_none()
    for config_file in opts.config:
        file_config = merge(file_config, from_file(config_file))

    # load options from cli
    cli_config = from_options(leftover_args)

    # merge file and cli config; cli config overrides file config
    input_config = merge(file_config, cli_config)

    # print both configs if requested
    if opts.inputs:
        pu.print("INPUT/CLI ARGS:", cli_config)
        pu.print("INPUT/CFG FILE:", file_config)

    if do_no_run and not opts.parsed:
        # if the user hasn't requested to print the parsed config
        # and we are not running the main program, we can exit here.
        sys.exit(0)

    # load configuration with node parsers
    parsed_config = merge(config_node, input_config)

    # check if all parameters are provided/resolved
    parsed_config = validate(parsed_config)

    # print it if requested
    if not (opts.quiet) or opts.parsed:
        pu.print("PARSE/ALL CFG:", parsed_config)

    if do_no_run:
        # we are not running because the user has requested to print
        # either the options, inputs, or parsed config.
        sys.exit(0)
    else:
        # we execute the main method and pass the parsed config to it
        return func(parsed_config, *args, **kwargs)


def cli(
    config_node_cls: Optional[Type[CT]] = None,
    print_fn: Optional[Callable] = None,
) -> Callable[
    [
        # this is a main method that takes as first input a parsed config
        Callable[Concatenate[CT, MP], RT]
    ],
    # the decorated method doesn't expect the parsed config as first input,
    # since that will be parsed from the command line
    Callable[MP, RT],
]:
    """
    TODO[lucas]: write doc
    """

    if config_node_cls is None:
        config_node = from_none()
        name = "<unnamed>"

    elif not (isclass(config_node_cls) and is_dataclass(config_node_cls)):
        msg = "`config_node` must be be decorated as a dataclass"
        raise ValueError(msg)
    else:
        config_node = from_dataclass(config_node_cls)
        name = config_node_cls.__name__

    def wrapper(func: Callable[Concatenate[CT, MP], RT]) -> Callable[MP, RT]:
        def wrapping(*args: MP.args, **kwargs: MP.kwargs) -> RT:
            # I could have used a functools.partial here, but defining
            # my own function instead allows me to provide nice typing
            # annotations for mypy.
            return wrap_main_method(
                func,
                name,
                config_node,
                print_fn,
                *args,
                **kwargs,
            )

        return wrapping

    return wrapper



import functools
import importlib
import importlib.util
import inspect
import itertools
from dataclasses import is_dataclass
from types import ModuleType
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from get_annotations import get_annotations
from omegaconf import DictConfig, OmegaConf

from .core import cast
from .utils import SpringsWarnings, clean_multiline


class InitLater(functools.partial):

    # inherits slots from functools.partial
    __slots__ = ("type_",)

    # must be explicitly defined to avoid mypy error
    type_: Union[Type[Any], None]

    def __new__(
        cls,
        func: Callable,
        _type_: Optional[type] = None,
        /,
        *args,
        **keywords,
    ):
        cl = super().__new__(cls, func, *args, **keywords)
        cl.type_ = _type_
        return cl

    def get_kw(self: "InitLater", *args, **kwargs) -> Any:
        """Shortcut for accessing parameters that have been
        provided to an InitLater object"""
        return self.keywords.get(*args, **kwargs)

    def pop_kw(self: "InitLater", *args, **kwargs) -> Any:
        """Shortcut for popping parameters that have been
        provided to an InitLater object"""
        return self.keywords.pop(*args, **kwargs)

    @staticmethod
    def _no_op():
        ...

    @classmethod
    def no_op(cls: Type["InitLater"]) -> "InitLater":
        """Create an init later that does nothing.
        Useful for when trying to instantiate from None."""
        return cls(cls._no_op)

    def __bool__(self: "InitLater") -> bool:
        # this allows us to check if the InitLater object
        # is "True" (that is, it is a real wrapped function),
        # or if it is "False" (it's a no-op)
        return self.func != self._no_op

    def __call__(self: "InitLater", /, *args, **kwargs):
        # recursively call deferred initialization if
        # we encounter another InitLater
        args = tuple(
            v() if isinstance(v, InitLater) else v
            for v in itertools.chain(self.args, args)
        )
        kwargs = {
            k: v() if isinstance(v, InitLater) else v
            for k, v in {**self.keywords, **kwargs}.items()
        }
        try:
            out = self.func(*args, **kwargs)

            # if type is provided, then we check if the object that
            # has been initialized here is of the expected type.
            # note that this only works for top-level init, and
            # does not recursively check.

            try:
                # We only type check if the type is provided
                if self.type_ is not None:
                    type_check_out = isinstance(out, self.type_)
                else:
                    type_check_out = True
            except TypeError:
                # Instance check fails if we can't check for type for
                # some reason, e.g. if self.type_ is None, self.type_ is a
                # Protocol or self._type is an instance of a class.
                #
                # If the instance check fails, we assume that the type check
                # went through, and return the object
                type_check_out = True

            if not type_check_out:
                msg = (
                    f"Initialized object `{out}` is not the right type: "
                    f"expected `{self.type_}`, got {type(out)}"
                )
                raise TypeError(msg)

            return out
        except Exception as e:
            ex_name = type(e).__name__
            fn_name = repr(self.func)
            msg = clean_multiline(
                f"""
                An error occurred while trying to initialize {fn_name}
                with arguments "{args}" and kwargs "{kwargs}":
                {ex_name}("{' '.join(map(str, e.args))}")
            """
            )
            raise RuntimeError(msg).with_traceback(e.__traceback__)


class Target:
    @staticmethod
    def _is_module(path: str) -> bool:
        try:
            spec = importlib.util.find_spec(path)
            return spec is not None
        except ModuleNotFoundError:
            return False

    @staticmethod
    def get_config_module(config: Any) -> Union[ModuleType, None]:
        """Returns the package of the type of config if the configuration
        is a structured config (that is, derived from a dataclass/attr.s,
        None otherwise"""
        if isinstance(config, DictConfig):
            config_type = OmegaConf.get_type(config)

            if is_dataclass(config_type):
                return inspect.getmodule(config_type)

        return None

    @classmethod
    def to_string(cls: Type["Target"], callable: Any) -> str:
        if inspect.isclass(callable):
            return f"{callable.__module__}.{callable.__name__}"
        elif inspect.ismethod(callable):
            return f"{cls.to_string(callable.__self__)}.{callable.__name__}"
        elif inspect.isfunction(callable):
            return f"{callable.__module__}.{callable.__name__}"
        else:
            raise TypeError(f"`{callable}` is not a callable")

    @classmethod
    def from_string(
        cls: Type["Target"], path: str, module: Optional[ModuleType] = None
    ) -> Any:
        """Returns a callable from a string path. If path is relative, then
        the module is used to resolve the path.

        Args:
            path (str): dot-separated path to the callable, e.g.
                "module.function"
            module (Optional[ModuleType]): module to use to resolve the path
                if the path is relative. If None, it is assumed that the
                path is absolute.
        """

        callable_ = None

        if path.startswith("."):
            if isinstance(module, ModuleType):
                path = f"{module.__package__}{path}"
            else:
                raise ImportError(f"Cannot resolve relative path {path}")

        if path.startswith("__main__."):
            # special case for any method that might have been defined in
            # the same file as the code entry point. For this case, we
            # first import `__main__`, and then look for the method there

            import __main__

            _, *method_name = path.rsplit(".")
            method_container = __main__

            while method_name:
                local_method_name = method_name.pop(0)

                if hasattr(method_container, local_method_name):
                    method_container = getattr(
                        method_container, local_method_name
                    )
                else:
                    # couldn't find the method
                    break

            # if we exhausted the method name, then we found the method
            callable_ = method_container if len(method_name) == 0 else None
        elif cls._is_module(path):
            return importlib.import_module(path)
        elif "." in path:
            m_name, c_name = path.rsplit(".", 1)
            if cls._is_module(m_name):
                container = importlib.import_module(m_name)
            else:
                container = cls.from_string(m_name)
            callable_ = getattr(container, c_name, None)
        else:
            # the if isinstance(..., dict) for builtin is so that mypy
            # does not complain, because __builtins__ is annotated as Any
            callable_ = globals().get(
                path,
                __builtins__.get(path, None)
                if isinstance(__builtins__, dict)
                else None,
            )

        if callable_ is None:
            raise ImportError(f"Cannot find callable at {path}")

        return callable_


InitT = TypeVar("InitT")
InitNestedT = TypeVar("InitNestedT")
CallableT = TypeVar("CallableT", bound=Callable)


class init(Generic[InitT, CallableT]):
    TARGET: str = "_target_"

    @classmethod
    def callable(
        cls: Type["init"],
        _config_: Optional[Any] = None,
        _type_: Optional[CallableT] = None,
        _recursive_: Optional[Any] = None,
        target: Optional[str] = None,
    ) -> CallableT:
        """Finds a callable function corresponding to the given `target`
        in the config. If no `target` is provided, it looks for `_target_`
        inside the config, raising a KeyError if not found.

        Args:
            config (Optional[DictConfig], optional): A DictConfig configuration
                to use. Defaults to None. config is expected to have at least
                one attribute named `_target_` which contains the path to
                the callable to instantiate.
            target (Optional[str], optional): The name of the target method
                to find. If not provided, it will look for `_target_` in the
                config. Defaults to None.
            _type_ (Optional[Type[InitT]], optional): Expected type of the
                object to instantiate. If provided, a type check is run on
                the instantiated object, raising a TypeError types don't match.
                If not provided, no type check is run. Defaults to None.

        Returns:
            A callable function that can be used to instantiate an object.
        """

        if target is not None and _config_ is not None:
            raise ValueError("Cannot specify both `target` and `config`")
        elif target is None:
            _config_ = cast(_config_)
            try:
                target = str(_config_[cls.TARGET])
            except KeyError:
                raise KeyError(
                    f"Config `{_config_}` has no `{cls.TARGET}` key!"
                )

        config_module = Target.get_config_module(_config_)
        return Target.from_string(target, config_module)

    @overload
    @classmethod
    def later(
        cls: Type["init"],
        _config_: Any,
        _type_: Type[InitT],
        _recursive_: bool = True,
        /,
        **kwargs: Any,
    ) -> Callable[..., InitT]:
        ...

    @overload
    @classmethod
    def later(
        cls: Type["init"],
        _config_: Optional[Any] = None,
        _type_: Optional[Type[InitT]] = None,
        _recursive_: bool = True,
        /,
        **kwargs: Any,
    ) -> Any:
        ...

    @classmethod
    def later(
        cls: Type["init"],
        _config_: Optional[Any] = None,
        _type_: Optional[Type[InitT]] = None,
        _recursive_: bool = True,
        /,
        **kwargs: Any,
    ) -> Callable[..., InitT]:
        """Returns a `functools.partial` object that can be used to initialize
        the object at a later time. The partial object can receive additional
        keyword arguments that are not initially available when calling
        `init.later`.

        Args:
            config (Optional[DictConfig], optional): A DictConfig configuration
                to use. Defaults to None. config is expected to have at least
                one attribute named `_target_` which contains the path to
                the callable to instantiate.
            _type_ (Optional[Type[InitT]], optional): Expected type of the
                object to instantiate. If provided, a type check is run on
                the instantiated object, raising a TypeError types don't match.
                If not provided, no type check is run. Defaults to None.
            kwargs (Dict[str, Any], optional): Keyword arguments to pass to the
                constructor for the object specified in the config. Defaults to
                an empty dictionary.

        Returns:
            An callable that returns an object of type `_type_`.
        """
        SpringsWarnings.missing_type(fn_name="init.later", type_=_type_)

        # if no config is provided, we return a function
        if _config_ is None:
            return InitLater.no_op()

        # ensure we are working with a ConfigDict
        config_node = cast(_config_)

        # try to get the target callable from either the config
        # or keyword arguments passed to the init function
        if cls.TARGET in config_node:
            fn = cls.callable(_config_=config_node, _type_=_type_)
        elif cls.TARGET in kwargs:
            fn = cls.callable(target=kwargs[cls.TARGET], _type_=_type_)
        else:
            raise ValueError(
                f"Cannot instantiate from `{config_node}` and "
                f"`{kwargs}`: `{cls.TARGET}` keyword missing"
            )

        def _recursive_init(
            param: Union[dict, DictConfig, None], type_: Optional[type] = None
        ) -> Any:
            """Given one of the parameters for the object that we are
            trying to instantiate, plus optionally its expected type, this
            function will try to instantiate the parameter if it is a nested
            config AND if _recursive_ is True.
            """
            must_recursive_init = (
                # we are operating in recursive mode
                _recursive_
                # the parameter is a nested config
                and isinstance(param, (DictConfig, dict))
                # the nested config has a TARGET property
                and cls.TARGET in param
            )
            if must_recursive_init:
                # we initialize recursively
                return cls.later(param, type_, _recursive_)
            else:
                # we simply return the parameter
                return param

        def _find_child_type(
            cls_: Union[None, type], attr_name: str
        ) -> Union[None, type]:
            """Given the type the object we are initializing is expected to be,
            as well as the name of the attribute we are about to pass to its
            initializer, this function figures out the expected type using
            a combination of class annotations and __init__ annotations.

            If the type the attribute should be cannot be determined, it
            simply returns None.

            An example:

            ```python
            class Test1:
                def __init__(self, a: int):
                    ...

            print(_find_child_type(Test1, "a")) # prints "int"

            class Test2:
                a: float

                def __init__(self, a):
                    ...

            print(_find_child_type(Test2, "a")) # prints "float"

            class Test3:
                def __init__(self, a):
                    ...

            print(_find_child_type(Test3, "a")) # prints "None"
            ```
            """

            if cls_ is None:
                # if the class is None, we don't perform any checks
                return None

            if attr_name in (parent_anns := get_annotations(cls_)):
                return parent_anns[attr_name]

            if attr_name in (
                parent_anns := inspect.getfullargspec(cls_).annotations
            ):
                return parent_anns[attr_name]

            # this is the case where we cannot resolve anything
            return None

        init_call_dict = {
            str(k): _recursive_init(
                param=v, type_=_find_child_type(cls_=_type_, attr_name=str(k))
            )
            for k, v in {**config_node, **kwargs}.items()
            if k != cls.TARGET
        }

        return InitLater(fn, _type_, **init_call_dict)

    @overload
    @classmethod
    def now(
        cls: Type["init"],
        _config_: Any,
        _type_: Type[InitT],
        _recursive_: bool = True,
        /,
        **kwargs: Any,
    ) -> InitT:
        ...

    @overload
    @classmethod
    def now(
        cls: Type["init"],
        _config_: Optional[Any] = None,
        _type_: Optional[Type[InitT]] = None,
        _recursive_: bool = True,
        /,
        **kwargs: Any,
    ) -> Any:
        ...

    @classmethod
    def now(
        cls: Type["init"],
        _config_: Optional[Any] = None,
        _type_: Optional[Type[InitT]] = None,
        _recursive_: bool = True,
        /,
        **kwargs: Any,
    ) -> InitT:
        """Instantiate a class or call a function from a configuration.

        Args:
            config (Optional[DictConfig], optional): A DictConfig configuration
                to use. Defaults to None. config is expected to have at least
                one attribute named `_target_` which contains the path to
                the callable to instantiate.
            _type_ (Optional[Type[InitT]], optional): Expected type of the
                object to instantiate. If provided, a type check is run on
                the instantiated object, raising a TypeError types don't match.
                If not provided, no type check is run. Defaults to None.
            kwargs (Dict[str, Any], optional): Keyword arguments to pass to the
                constructor for the object specified in the config. Defaults to
                an empty dictionary.

        Returns:
            An object of type `_type_`.
        """
        SpringsWarnings.missing_type(fn_name="init.now", type_=_type_)

        # notice the use of non-keyword arguments here for config,
        # _type_, and _recursive_. This is because `later` has a `/`
        # in its signature, which is required to ensure that **kwargs
        # works well with type annotations.
        init_call: Callable[..., InitT] = cls.later(
            _config_, _type_, _recursive_, **kwargs
        )
        return init_call()

    """Convenience shortcut for `init.now`"""
    __new__: Callable[..., InitT] = now  # type: ignore



from dataclasses import is_dataclass
from typing import (
    Callable,
    Dict,
    Literal,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

T = TypeVar("T")


class NicknameRegistry:
    __registry__: Dict[str, Type] = {}

    @classmethod
    def add(cls, name: str) -> Callable[[Type[T]], Type[T]]:
        """Save a configuration with a nickname for easy reuse."""

        def add_to_registry(fn: Type[T]) -> Type[T]:
            if not is_dataclass(fn):
                raise ValueError(f"{fn} must be a dataclass")

            if name in cls.__registry__:
                raise ValueError(f"{name} is already registered")
            cls.__registry__[name] = fn
            return fn

        return add_to_registry

    @overload
    @classmethod
    def get(cls, name: str, raise_if_missing: Literal[True] = True) -> Type[T]:
        ...

    @overload
    @classmethod
    def get(
        cls, name: str, raise_if_missing: Literal[False] = False
    ) -> Union[Type, None]:
        ...

    @classmethod
    def get(
        cls, name: str, raise_if_missing: Literal[True, False] = False
    ) -> Union[Type, None]:
        if raise_if_missing and name not in cls.__registry__:
            raise ValueError(f"{name} is not registered as a nickname")
        return cls.__registry__.get(name, None)

    @classmethod
    def all(cls) -> Sequence[Tuple[str, str]]:
        return [
            (name, str(config.__name__))
            for name, config in cls.__registry__.items()
        ]



import importlib.metadata
from dataclasses import dataclass, field
from typing import Callable, Optional, Type, TypeVar

from omegaconf import II, MISSING, SI

from .commandline import cli
from .core import (
    cast,
    from_dataclass,
    from_dict,
    from_file,
    from_none,
    from_options,
    from_string,
    merge,
    to_dict,
    to_yaml,
    validate,
)
from .flexyclasses import flexy_field, flexyclass, make_flexy
from .initialize import Target, init
from .logging import configure_logging
from .nicknames import NicknameRegistry
from .resolvers import all_resolvers, register
from .traversal import traverse
from .types import get_type
from .utils import SpringsWarnings

T = TypeVar("T")


try:
    # package has been installed, so it has a version number
    # from pyproject.toml
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    # package hasn't been installed, so set version to "dev"
    __version__ = "dev"


def toggle_warnings(value: Optional[bool] = None):
    """Shortcut for springs.utils.SpringsWarnings.toggle"""
    SpringsWarnings.toggle(value)


def make_target(c: Callable) -> str:
    """Shortcut for springs.initialize.Target.to_string"""
    return Target.to_string(c)


def nickname(name: str) -> Callable[[Type[T]], Type[T]]:
    """Shortcut for springs.nicknames.NicknameRegistry.add"""
    return NicknameRegistry.add(name)


__all__ = [
    "all_resolvers",
    "cast",
    "cli",
    "configure_logging",
    "dataclass",
    "field",
    "flexyclass",
    "flexy_field",
    "from_dataclass",
    "from_dict",
    "from_file",
    "from_none",
    "from_options",
    "from_string",
    "get_type",
    "II",
    "init",
    "make_flexy",
    "make_target",
    "merge",
    "MISSING",
    "nickname",
    "register",
    "SI",
    "Target",
    "toggle_warnings",
    "to_dict",
    "to_yaml",
    "traverse",
    "validate",
]



import sys
import types
import warnings
from collections import abc
from dataclasses import fields, is_dataclass
from typing import (
    Any,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

from omegaconf import DictConfig, ListConfig, OmegaConf

try:
    from omegaconf._utils import get_type_hint as _get_type_hint

    GET_TYPE_HINT_AVAILABLE = True
except ImportError:
    GET_TYPE_HINT_AVAILABLE = False


def get_type_hint(obj: Any, key: Any = None) -> Optional[Type[Any]]:
    if GET_TYPE_HINT_AVAILABLE:
        return _get_type_hint(obj, key)
    else:
        warnings.warn(
            "get_type_hint could not be imported from "
            "omegaconf._utils, falling back to OmegaConf.get_type"
        )
        return OmegaConf.get_type(obj, key)


NoneType: Type[None] = type(None)


def is_union_annotation(type_: Any) -> bool:
    """Check wether `type_` is equivalent to `typing.Union[T, ...]`

    From https://github.com/omry/omegaconf/blob/e95c2c76d2545a844794682108ded57fbf98f042/omegaconf/_utils.py#L195"""  # noqa: E501
    if sys.version_info >= (3, 10):
        if isinstance(type_, types.UnionType):
            return True
    return getattr(type_, "__origin__", None) is Union


def resolve_optional(type_: Any) -> Tuple[bool, Any]:
    """Check whether `type_` is equivalent to `typing.Optional[T]`
    for some T.

    From https://github.com/omry/omegaconf/blob/e95c2c76d2545a844794682108ded57fbf98f042/omegaconf/_utils.py#L202"""  # noqa: E501
    if is_union_annotation(type_):
        args = type_.__args__
        if NoneType in args:
            optional = True
            args = tuple(a for a in args if a is not NoneType)
        else:
            optional = False
        if len(args) == 1:
            return optional, args[0]
        elif len(args) >= 2:
            return optional, Union[args]  # type: ignore
        else:
            assert False

    if type_ is Any:
        return True, Any

    if type_ in (None, NoneType):
        return True, NoneType

    return False, type_


class MappingType(NamedTuple):
    key: Any
    val: Any


def resolve_mapping(type_: Any) -> Union[None, MappingType]:
    origin = get_origin(type_)

    if origin is not None and issubclass(origin, abc.Mapping):
        return MappingType(*get_args(type_))

    return None


def resolve_sequence(type_: Any) -> Union[None, MappingType]:
    origin = get_origin(type_)

    if origin is not None and issubclass(origin, abc.Sequence):
        return MappingType(*get_args(type_))

    return None


def safe_select(config: DictConfig, key: str, interpolate: bool = True) -> Any:
    """Selects a key from a config, but returns None if the key
    is missing or the key resolution fails."""

    if (
        key in config
        and OmegaConf.is_interpolation(config, key)
        and not interpolate
    ):
        return OmegaConf.to_container(config).get(key, None)  # type: ignore

    if (key not in config) or OmegaConf.is_missing(config, key):
        return None
    elif OmegaConf.is_interpolation(config, key):
        if interpolate:
            return OmegaConf.select(
                cfg=config, key=key, throw_on_resolution_failure=False
            )
        else:
            di: dict = OmegaConf.to_container(config)  # type: ignore
            return di.get(key, None)
    else:
        return OmegaConf.select(cfg=config, key=key)


def get_type(
    config_node: Union[DictConfig, ListConfig],
    key: Optional[Union[int, str]] = None,
) -> Union[type, None]:
    """Tries to infer the type of a config node key. Reurns None if
    the type cannot be inferred."""

    if not isinstance(config_node, (DictConfig, ListConfig)):
        raise ValueError(
            "Expected a DictConfig or ListConfig object, "
            f"got {type(config_node)} instead"
        )

    if key is None:
        return OmegaConf.get_type(config_node)

    if isinstance(config_node, DictConfig):
        typ_ = OmegaConf.get_type(config_node, str(key))

        if typ_ is not None:
            return typ_

        node_type = OmegaConf.get_type(config_node)
        node_type_hint = get_type_hint(config_node)

        if is_dataclass(node_type):
            for field in fields(node_type):
                if field.name == key:
                    typ_ = field.type
                    break
            _, typ_ = resolve_optional(typ_)

        elif resolved_node_type_hint := resolve_mapping(node_type_hint):
            # we return the default type that was given to all type hints
            typ_ = (
                resolved_node_type_hint.val
                if resolved_node_type_hint is not None
                else None
            )

        return typ_

    elif isinstance(config_node, ListConfig):
        if not isinstance(key, int):
            raise ValueError(f"Expected an int key, got {type(key)} instead")

        if 0 <= key < len(config_node):
            # you asked for type of an existing element
            return OmegaConf.get_type(config_node[key])
        else:
            # type of a non-existing element, so we rely on type hint
            # if available
            node_type_hint = get_type_hint(config_node)
            resolved_node_type_hint = resolve_sequence(node_type_hint)

            # gets around the fact that we might not be able resolve
            return (
                resolved_node_type_hint.val
                if resolved_node_type_hint is not None
                else None
            )
    else:
        raise ValueError(
            "Expected a DictConfig or ListConfig object, "
            f"got {type(config_node)} instead"
        )



from collections import abc
from copy import deepcopy
from dataclasses import is_dataclass
from pathlib import Path
from typing import Any, Dict, Sequence, Union

from omegaconf import DictConfig, ListConfig, OmegaConf
from omegaconf.omegaconf import DictKeyType

from .flexyclasses import unlock_all_flexyclasses
from .traversal import traverse
from .types import get_type, safe_select


@unlock_all_flexyclasses
def cast(config: Any, copy: bool = False) -> DictConfig:
    if is_dataclass(config):
        parsed_config = from_dataclass(config)
    elif isinstance(config, dict):
        parsed_config = from_dict(config)
    elif isinstance(config, str):
        parsed_config = from_string(config)
    elif config is None:
        parsed_config = from_none(config)
    elif isinstance(config, DictConfig):
        parsed_config = deepcopy(config) if copy else config
    elif isinstance(config, Path):
        parsed_config = from_file(config)
    else:
        raise TypeError(f"Cannot cast `{type(config)}` to DictConfig")

    return parsed_config


@unlock_all_flexyclasses
def from_none(*args: Any, **kwargs: Any) -> DictConfig:
    """Returns an empty dict config"""
    return OmegaConf.create()


@unlock_all_flexyclasses
def from_dataclass(config: Any) -> DictConfig:
    """Cast a dataclass to a structured omega config"""
    if config is None:
        return from_none()

    if not is_dataclass(config):
        raise TypeError(f"`{config}` is not a dataclass!")

    parsed_config = OmegaConf.structured(config)

    if not isinstance(parsed_config, DictConfig):
        raise TypeError(f"Cannot create dict config from `{config}`")
    return parsed_config


@unlock_all_flexyclasses
def from_dict(
    config: Union[Dict[DictKeyType, Any], Dict[str, Any]]
) -> DictConfig:
    """Create a config from a dict"""
    if not isinstance(config, dict):
        raise TypeError(f"`{config}` is not a dict!")

    parsed_config = OmegaConf.create(config)

    if not isinstance(parsed_config, DictConfig):
        raise ValueError(f"Config `{config}` is not a DictConfig!")
    return OmegaConf.create(config)


@unlock_all_flexyclasses
def from_string(config: str) -> DictConfig:
    """Load a config from a string"""
    if not isinstance(config, str):
        raise TypeError(f"`{config}` is not a string!")

    parsed_config = OmegaConf.create(config)

    if not isinstance(parsed_config, DictConfig):
        raise ValueError(f"Config `{config}` is not a DictConfig!")
    return parsed_config


@unlock_all_flexyclasses
def from_file(path: Union[str, Path]) -> DictConfig:
    """Load a config from a file"""
    path = Path(path)

    if not path.exists():
        raise FileNotFoundError(f"Cannot file configuration at {path}")
    config = OmegaConf.load(path)

    if not isinstance(config, DictConfig):
        raise ValueError(f"Config loaded from {path} is not a DictConfig!")

    return config


@unlock_all_flexyclasses
def from_options(opts: Sequence[str]) -> DictConfig:
    """Create a config from a list of options"""
    if not isinstance(opts, abc.Sequence) or not all(
        isinstance(o, str) for o in opts
    ):
        raise TypeError(f"`{opts}` is not a list of strings!")

    config = OmegaConf.from_dotlist(list(opts))
    if not isinstance(config, DictConfig):
        raise TypeError(f"input is not a sequence of strings, but `{opts}")
    return config


def to_yaml(config: Any) -> str:
    """Convert a omegaconf config to a YAML string"""
    if not isinstance(config, DictConfig):
        config = from_dataclass(config)
    return OmegaConf.to_yaml(config)


def to_dict(config: Any) -> Dict[DictKeyType, Any]:
    """Convert a omegaconf config to a Python primitive type"""
    if is_dataclass(config):
        config = from_dataclass(config)
    container = OmegaConf.to_container(config)

    if not isinstance(container, dict):
        raise TypeError(f"`{container}` is not a dict!")

    return container


########################################


def validate(config_node: Any) -> DictConfig:
    """Check if all attributes are resolve and not missing"""

    if not isinstance(config_node, DictConfig):
        raise TypeError(f"`{config_node}` is not a DictConfig!")

    for spec in traverse(config_node):
        if spec.key is None:
            raise RuntimeError(
                "You should not be here! Something went "
                "wrong in the core Springs library."
            )

        if OmegaConf.is_missing(spec.parent, spec.key):
            raise ValueError(f"Missing value for `{spec.path}`")
        if OmegaConf.is_interpolation(spec.parent, spec.key):
            try:
                getattr(spec.parent, str(spec.key))
            except Exception:
                raise ValueError(
                    f"Interpolation for `{spec.path}` " "not resolved"
                )

    config_node = deepcopy(config_node)
    OmegaConf.resolve(config_node)
    return config_node


def _pre_merge_fix_type_mismatches(
    merge_into: DictConfig, merge_from: DictConfig
) -> None:
    """Sometimes, when merging, new nodes or attributes appear in the
    configuration we are merging from; we need to make sure these nodes
    are properly initialized in the configuration we are merging into,
    or else merging will fail."""

    node_it = (
        range(len(merge_from))
        if isinstance(merge_from, ListConfig)
        else tuple(merge_from.keys())
    )

    for key in node_it:
        key = str(key)  # linter gets confused without this casting

        merge_from_value = safe_select(merge_from, key, interpolate=False)
        merge_into_value = safe_select(merge_into, key, interpolate=False)
        merge_into_expected_type = get_type(merge_into, key)

        if isinstance(merge_from_value, DictConfig):
            if isinstance(merge_into_value, DictConfig):
                # both configs have nodes at this location, so we need to
                # recursively initialize new nodes down in the tree.

                _pre_merge_fix_type_mismatches(
                    merge_into_value, merge_from_value
                )

            elif merge_into_expected_type:
                if is_dataclass(merge_into_expected_type):
                    # the merge_into node is not a configuration, but it could
                    # be one, since its type is a dataclass. Therefore, we
                    # first initialize this its node with an empty dataclass,
                    # which will then cause no issue when merging.
                    merge_into_value = from_dataclass(merge_into_expected_type)

                    # merge = False ensure replacement
                    OmegaConf.update(
                        cfg=merge_into,
                        key=key,
                        value=merge_into_value,
                        merge=False,
                    )
                    # the cast is necessary to make sure that,
                    # after the update, all flexyclasses are unlocked
                    merge_into = cast(merge_into)

                    # now that we have a proper dataclass here, we again
                    # recursively see if there are any new nodes to initialize.
                    _pre_merge_fix_type_mismatches(
                        merge_into_value, merge_from_value
                    )


def _pre_merge_override_interpolations(
    merge_into: DictConfig, merge_from: DictConfig
) -> None:
    """Merge interpolations from merge_from into merge_into"""

    for key in list(merge_from.keys()):
        key = str(key)  # linter gets confused without this casting

        merge_from_value = safe_select(merge_from, key, interpolate=False)
        merge_into_value = safe_select(merge_into, key, interpolate=False)

        if OmegaConf.is_interpolation(merge_from, key):
            # merge = False ensure replacement
            OmegaConf.update(
                cfg=merge_into, key=key, value=merge_from_value, merge=False
            )
            # the cast is necessary to make sure that,
            # after the update, all flexyclasses are unlocked
            merge_into = cast(merge_into)
            delattr(merge_from, key)

        elif isinstance(merge_from_value, DictConfig) and isinstance(
            merge_into_value, DictConfig
        ):
            _pre_merge_override_interpolations(
                merge_into_value, merge_from_value
            )


def merge(*configs: Any) -> DictConfig:
    """Merges multiple configurations into one."""

    if not configs:
        # no configs were provided, return an empty config
        return from_none()

    if not all(isinstance(c, DictConfig) for c in configs):
        raise TypeError(f"`{configs}` is not a list of DictConfigs!")

    # make sure all configs are DictConfigs
    merged_config, *other_configs = (
        cast(config, copy=True) for config in configs
    )

    # do the actual merging; this will also check if types are compatible
    for other_config in other_configs:

        _pre_merge_fix_type_mismatches(merged_config, other_config)
        _pre_merge_override_interpolations(merged_config, other_config)

        # ignoring type mypy check since merge_config and other_config are
        # DictConfig, so the return type is always DictConfig
        merged_config: DictConfig = OmegaConf.merge(  # type: ignore
            merged_config, other_config
        )

        #  raise error if we end up with something that is not a DictConfig
        if not isinstance(merged_config, DictConfig):
            raise TypeError(
                f"While merging {configs}, the resulting config is"
                f" {type(merged_config)} instead of DictConfig."
            )

    return cast(merged_config)



import os
import re
import shutil
import warnings
from ast import literal_eval
from textwrap import dedent
from typing import Any, Callable, Optional, Type, Union

import yaml
from omegaconf import DictConfig, OmegaConf


def clean_multiline(string: str) -> str:
    string = dedent(string.strip())
    string = re.sub(r"\s*\n\s*", " ", string)
    return string


class PrintUtils:
    """A few utilities to make printing easier"""

    def __init__(
        self,
        indent_step: int = 2,
        indent_char: str = " ",
        separator_char: str = "-",
        terminal_width: Optional[int] = None,
        print_fn: Optional[Callable[..., None]] = None,
    ) -> None:
        self.indent_step = indent_step
        self.indent_char = indent_char
        self.separator_char = separator_char
        self.terminal_width = (
            terminal_width or shutil.get_terminal_size().columns
        )
        self.print_fn = print_fn or print
        self.has_printed_first_separator = False
        self.is_default_print = self.print_fn is print

    def print(
        self,
        *lines: Union[str, dict, DictConfig],
        level: int = 0,
        level_up: Optional[int] = None,
    ):

        level_up = level_up or float("inf")  # type: ignore

        content = ""
        if self.is_default_print:
            if not self.has_printed_first_separator:
                content += self.separator()
            self.has_printed_first_separator = True

        for line in lines:
            if not line:
                # backtrack the new line you just added from previous
                # argument if the current argument evaluates to none, like
                # an empty string, or an empty dict
                content = f"{content.rstrip()} "

            if isinstance(line, DictConfig):
                # we ignore type bc we are ok with line being any of
                # the types returned by omegaconf
                line = OmegaConf.to_container(line)  # type: ignore

            if isinstance(line, dict):
                content += f"{self.to_yaml(line, level)}\n"
            else:
                content += f"{self.indent(str(line), level)}\n"

            level += 1 if level < level_up else 0  # type: ignore

        if self.is_default_print:
            content += f"{self.separator()}\n"

        self.print_fn(content.rstrip())  # type: ignore

    def indent(self, line: str, level: int) -> str:
        indent = self.indent_char * level * self.indent_step
        return indent + line

    def separator(self, level: int = 0) -> str:
        separator = self.separator_char * self.terminal_width
        return self.indent(separator, level=level)

    def to_yaml(self, content: dict, level: int = 0) -> str:
        if not content:
            # this is in case the dict/list is empty
            return str(content)

        out = yaml.dump(
            content,
            indent=self.indent_step,
            width=self.terminal_width,
            default_flow_style=False,
        )
        out = "\n".join(
            self.indent(ln, level) for ln in out.strip().split("\n")
        ).rstrip()
        return out


class SpringsWarnings:
    _WARNINGS = literal_eval(os.environ.get("SPRINGS_WARNINGS", "True"))

    @classmethod
    def toggle(cls, value: Optional[bool] = None):
        if value is None:
            value = not cls._WARNINGS
        cls._WARNINGS = value

    @classmethod
    def _warn(
        cls: Type["SpringsWarnings"],
        message: str,
        category: Type[Warning],
        stacklevel: int = 2,
    ):
        if cls._WARNINGS:
            warnings.warn(message, category, stacklevel=stacklevel)

    @classmethod
    def missing_type(cls: Type["SpringsWarnings"], fn_name: str, type_: Any):
        if type_ is None:
            cls._warn(
                clean_multiline(
                    f"""It is strongly recommended to provide a _type_ argument
                        to `{fn_name}`. This ensures that the correct type is
                        annotated as the return value. Further, it performs
                        type checking on the initialized object."""
                ),
                category=UserWarning,
            )

    @classmethod
    def flexyclass(cls: Type["SpringsWarnings"]):
        cls._warn(
            clean_multiline(
                """Decorating with `flexyclass` is discouraged because it
                   does not play nicely with mypy, resulting in incorrect
                   type annotations. Instead, consider decorating with
                   `@springs.dataclass` first, and then decorating
                   `@springs.make_flexy` on the resulting class."""
            ),
            category=UserWarning,
        )

    @classmethod
    def flexyfield(cls: Type["SpringsWarnings"]):
        cls._warn(
            clean_multiline(
                """`flexy_field` is provided as a convenience for passing
                values that are not explicitly annotated for a flexyclass;
                however, it is not recommended because it not very pythonic.
                Instead, creating an flexyclass that inherits from the current
                one with the desired field and default value. Non explicit
                values are expected to be provided at runtime. """
            ),
            category=UserWarning,
        )

    @classmethod
    def argument(cls: Type["SpringsWarnings"], arg_name: str, obj_name: str):
        cls._warn(
            f"'{arg_name}' was provided to `{obj_name}`, but it is ignored",
            category=SyntaxWarning,
        )



from datetime import datetime, timezone
from pathlib import Path
from typing import Callable, Optional, Sequence, TypeVar, Union

from omegaconf import DictConfig, OmegaConf
from omegaconf.basecontainer import BaseContainer
from yaml.error import MarkedYAMLError

from springs.core import from_dataclass, from_options, merge

from .nicknames import NicknameRegistry

T = TypeVar("T")


def register(
    name: str, use_cache: bool = False
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def _register(func: Callable[..., T]) -> Callable[..., T]:

        # will raise an error if the resolver is already registered
        OmegaConf.register_new_resolver(
            name=name, resolver=func, use_cache=use_cache, replace=False
        )
        return func

    return _register


def all_resolvers() -> Sequence[str]:
    return [str(k) for k in BaseContainer._resolvers.keys()]


@register("sp.fullpath")
def get_full_path(path: str) -> str:
    """Resolve all implicit and relative path components
    to give an absolute path to a file or directory"""
    return str(Path(path).resolve().absolute())


@register("sp.timestamp")
def get_timestamp(fmt: Optional[str] = None) -> str:
    """Returns a timestamp in the format provided; if not provided, use
    year-month-day_hour-minute-second."""

    fmt = fmt or "%Y-%m-%d_%H-%M-%S"
    return datetime.now(tz=timezone.utc).strftime(fmt)


@register("sp.from_node")
def from_node(
    node_or_nickname: Union[DictConfig, str], *values: str
) -> DictConfig:
    """Instantiates a node from another node of a nickname to a config.

    Example:

    ```python
    import springs as sp

    cfg = sp.from_dict({
        'train': {
            'data': {'path': '/train'},
            'name': 'train',
            'bs': 32
        },
            'test': '${sp.from_node: ${train}, "name=test", "data.path=/test"}'
    })
    print(sp.to_dict(sp.validate(cfg)))

    # Prints the following
    # {'test': {'data': {'path': '/test'}, 'name': 'test', 'bs': 32},
    #  'train': {'data': {'path': '/train'}, 'name': 'train', 'bs': 32}}
    ```

    Args:
        node_or_nickname: The node or nickname to instantiate from;
            if a nickname, the node is looked up in the registry;
            if node, it must be a valid omegaconf DictConfig.
        values: The overrides to pass to the node; they must be in the
            format of command line arguments, e.g. "path.to.key=value".
    """

    if isinstance(node_or_nickname, str):
        node = from_dataclass(NicknameRegistry.get(node_or_nickname))
    elif isinstance(node_or_nickname, DictConfig):
        node = node_or_nickname
    else:
        raise TypeError(
            "node must be a reference to another node in the config or "
            "the nickname assigned to a structured config, not"
            f"not {type(node_or_nickname)}"
        )

    if not values:
        # no replacement to do!
        return node

    try:
        if not all(isinstance(v, str) for v in values):
            raise TypeError
        replace = from_options(values)
    except (MarkedYAMLError, TypeError) as e:
        raise ValueError(
            "sp.from_node overrides must be in the path.to.key=value format, "
            f"not {' '.join(values)}"
        ) from e

    # do the merging
    new_node = merge(node, replace)

    return new_node
